//virtual multichannel AY-3-8910
//all channels having tone/noise/envelope counters, with normal registers array
//based on my old code from 2004
//heavily reworked, optimized, and cut down to the synth needs



#ifndef CHIP_CHANNELS_ALL
#define CHIP_CHANNELS_ALL	16
#endif



struct chipToneStruct {

	int count;
	int div;
	int state;
	int enable;

};

struct chipNoiseStruct {

	int ptr;
	int count;
	int div;
	int state;
	int enable;

};

struct chipEnvStruct {

	const int *table;
	int ptr;
	int count;
	int div;
	int dac;
	int up;
	int enable;

};

struct chipChannelStruct {

	chipToneStruct tone;
	chipNoiseStruct noise;
	chipEnvStruct env;

	int volume;
	int out;

};

struct chipStruct {

	chipChannelStruct ch[CHIP_CHANNELS_ALL];

	const int *table;

	int div;

};



//volume tables, default ones from UnrealSpeccy

const int table_volume_ay[32]={
	0x0000,0x0000,0x0340,0x0340,0x04C0,0x04C0,0x06F2,0x06F2,
	0x0A44,0x0A44,0x0F13,0x0F13,0x1510,0x1510,0x227E,0x227E,
	0x289F,0x289F,0x414E,0x414E,0x5B21,0x5B21,0x7258,0x7258,
	0x905E,0x905E,0xB550,0xB550,0xD7A0,0xD7A0,0xFFFF,0xFFFF
};

const int table_volume_ym[32]={
	0x0000,0x0000,0x00EF,0x01D0,0x0290,0x032A,0x03EE,0x04D2,
	0x0611,0x0782,0x0912,0x0A36,0x0C31,0x0EB6,0x1130,0x13A0,
	0x1751,0x1BF5,0x20E2,0x2594,0x2CA1,0x357F,0x3E45,0x475E,
	0x5502,0x6620,0x7730,0x8844,0xA1D2,0xC102,0xE0A2,0xFFFF
};

//precalculated envelope waveforms

const int table_envelope[64*4]={
	0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,	//saw up
	0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,

	31,30,29,28,27,26,25,24,23,22,21,20,19,18,17,16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1,0,	//saw down
	31,30,29,28,27,26,25,24,23,22,21,20,19,18,17,16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1,0,

	0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,	//tri up/down
	31,30,29,28,27,26,25,24,23,22,21,20,19,18,17,16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1,0,

	31,30,29,28,27,26,25,24,23,22,21,20,19,18,17,16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1,0,	//tri down/up
	0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
};

//array for prerendered LFSR sequence

unsigned char table_noise[128*1024];



void ay_init(chipStruct *ay,int type)
{
	int i,n_di,n_reg,n_qcc,n_state,chn;

	memset(ay,0,sizeof(chipStruct));
	
	n_reg=0x0ffff;
	n_qcc=0;
	n_state=0;

	for(i=0;i<sizeof(table_noise);++i)
	{
		table_noise[i]=n_state;

		n_di=(n_qcc^((n_reg>>13)&1))^1;
		n_qcc=(n_reg>>15)&1;
		n_state=n_qcc;
		n_reg=(n_reg<<1)|n_di;
	}

	if(!type) ay->table=table_volume_ay; else ay->table=table_volume_ym;

	for(chn=0;chn<CHIP_CHANNELS_ALL;++chn) ay->ch[chn].env.table=table_envelope;
}



inline void ay_write(chipStruct *ay,int chn,unsigned char reg,unsigned char value)
{
	chipChannelStruct *ch;

	ch=&ay->ch[chn];

	switch(reg)
	{
	case 0:
		ch->tone.div=(ch->tone.div&0x0f00)|value;
		break;

	case 1:
		ch->tone.div=(ch->tone.div&0x00ff)|(value<<8);
		break;

	case 6:
		ch->noise.div=value&0x1f;
		break;

	case 7:
		ch->tone.enable=value&1;
		ch->noise.enable=(value>>3)&1;
		break;

	case 8:
		ch->volume=(value&0x0f)<<1;
		ch->env.enable=value&0x10;
		break;

	case 11:
		ch->env.div=(ch->env.div&0xff00)|value;
		break;

	case 12:
		ch->env.div=(ch->env.div&0x00ff)|(value<<8);
		break;

	case 13: //R13 write also resets the envelope counter

		switch(value&15)
		{
		case 0x08: ch->env.table=&table_envelope[1*64]; break;	//saw down
		case 0x0a: ch->env.table=&table_envelope[3*64]; break;	//tri down/up
		case 0x0c: ch->env.table=&table_envelope[0*64]; break;	//saw up
		case 0x0e: ch->env.table=&table_envelope[2*64]; break;	//tri up/down
		}

		ch->env.count=0;
		ch->env.ptr=0;

		break;
	}

	if(reg<2) if(!ch->tone.div) ch->tone.div=1;	//period 0 the same as period 1 for tone channels
}



inline void ay_tick(chipStruct *ay)
{
	chipChannelStruct *ch;
	int chn,dac;
	bool change;

	for(chn=0;chn<CHIP_CHANNELS_ALL;++chn)
	{
		ch=&ay->ch[chn];

		change=false;
		
		if(!ay->div)
		{
			//tone channel

			if(ch->tone.count>=ch->tone.div)
			{
				ch->tone.count=0;
				ch->tone.state^=1;
				
				change=true;
			}

			++ch->tone.count;

			//noise channel

			if(ch->noise.count>=ch->noise.div)
			{
				ch->noise.count=0;
				ch->noise.state=table_noise[ch->noise.ptr&131071];

				++ch->noise.ptr;
				
				change=true;
			}

			++ch->noise.count;
		}

		if(ch->env.count>=ch->env.div)
		{
			ch->env.count=0;

			ch->env.dac=ch->env.table[ch->env.ptr&63];

			++ch->env.ptr;
			
			change=true;
		}

		++ch->env.count;

		if(change)	//only calculate output level when it actually changed
		{
			//logic mixer

			if((ch->tone.state|ch->tone.enable)&(ch->noise.state|ch->noise.enable))
			{
				if(!ch->env.enable) dac=ch->volume; else dac=ch->env.dac;
			}
			else
			{
				dac=0;
			}

			//output level

			ch->out=ay->table[dac];
		}
	}

	ay->div^=1;
}
